--
-- The 'Mersenne Twister' random number generator
--  http://www.math.keio.ac.jp/matumoto/emt.html
--
-- Translated from C to Euphoria by Brendon Sly
--  (bwsly@infoscience.otago.ac.nz)
--
-- Version 1.1 of the Sly Translation
-- Modified July 2001 for speed increase,
-- by Rod Jackson (rodjackson@bigfoot.com)
--

include machine.e

constant N = 624, M = 397
constant MATRIX_A = #9908B0DF
constant UPPER_MASK = #80000000, LOWER_MASK = #7FFFFFFF

constant TEMPERING_MASK_B = #9D2C5680, TEMPERING_MASK_C = #EFC60000

constant MAX_4_BYTE_UINT = power (2,32) -- 1 isn't subtracted because it'll
                                        -- just be added back later....

sequence mt -- the actual twister
mt = repeat(0, N)
integer mti -- the current position within the twister .. 'index'
mti = N+1

function shift_l(atom val, integer num_to_shift)
 return (floor(val * power(2, num_to_shift)))
end function

function shift_r(atom val, integer num_to_shift)
 return (floor(val / power(2, num_to_shift)))
end function



atom anded_result
function bit_and(atom uint32_1, atom uint32_2)
 
 uint32_1 = remainder (uint32_1, MAX_4_BYTE_UINT)
 uint32_2 = remainder (uint32_2, MAX_4_BYTE_UINT)

 anded_result = and_bits (uint32_1, uint32_2)

 if (anded_result < 0) then
    anded_result = MAX_4_BYTE_UINT + anded_result
 end if

 return anded_result

end function

 ----------

atom ored_result
function bit_or(atom uint32_1, atom uint32_2)
 
 uint32_1 = remainder (uint32_1, MAX_4_BYTE_UINT)
 uint32_2 = remainder (uint32_2, MAX_4_BYTE_UINT)

 ored_result = or_bits (uint32_1, uint32_2)

 if (ored_result < 0) then
    ored_result = MAX_4_BYTE_UINT + ored_result
 end if
 
 return ored_result

end function

 ----------

atom xored_result
function bit_xor(atom uint32_1, atom uint32_2)
 
 uint32_1 = remainder (uint32_1, MAX_4_BYTE_UINT)
 uint32_2 = remainder (uint32_2, MAX_4_BYTE_UINT)

 xored_result = xor_bits (uint32_1, uint32_2)

 if (xored_result < 0) then
    xored_result = MAX_4_BYTE_UINT + xored_result
 end if

 return xored_result

end function

-- sgenrand 'seeds' the twister .. it's called by 'genrand' if needed
procedure sgenrand(atom seed)
 if (seed < 0) then
  seed = 0
 end if
 for i = 1 to N do
  mt[i] = bit_and(seed, #FFFF0000)
  seed = remainder ((69069 * seed + 1), MAX_4_BYTE_UINT)
  mt[i] = bit_or(mt[i], shift_r( bit_and(seed, #FFFF0000), 16) )
  seed = remainder ((69069 * seed + 1), MAX_4_BYTE_UINT)
 end for
 mti = N
end procedure

-- seed the twister with a sequence (of integers or atoms) instead
procedure lsgenrand(sequence seed_array)
 if (length(seed_array) = N) then
  for i = 1 to N do
   if not (sequence(seed_array[i])) then
    mt[i] = seed_array[i]
   else
    mt[i] = 0
   end if
  end for
  mti = N
 end if
end procedure

function genrand()
 atom y
 sequence mag01 

 y = 0
 mag01 = {#0, MATRIX_A}
 
 if (mti >= N) then
  
  if (mti = N+1) then
   -- here is where you'd add your default MT intialisation seed
   --  .. you could use a call to rand() or grab the computer time
   sgenrand(4357) -- 4357 is the standard MT test seed 
  end if

  for kk = 1 to (N-M) do
   y = bit_or( bit_and(mt[kk], UPPER_MASK), bit_and(mt[kk+1], LOWER_MASK) )
   mt[kk] = bit_xor( bit_xor(mt[kk+M], shift_r(y, 1)) , mag01[ bit_and(y, 1) + 1 ] )
  end for
  
  for kk = (N-M)+1 to N-1 do
   y = bit_or( bit_and(mt[kk], UPPER_MASK), bit_and(mt[kk+1], LOWER_MASK) )
   mt[kk] = bit_xor( bit_xor(mt[kk+(M-N)], shift_r(y, 1)) , mag01[ bit_and(y, 1) + 1 ] )
  end for
        
  y = bit_or( bit_and(mt[N], UPPER_MASK), bit_and(mt[1], LOWER_MASK) )     
  mt[N] = bit_xor( bit_xor(mt[M], shift_r(y, 1)) , mag01[ bit_and(y, 1) + 1 ] )

  mti = 0
 end if

 mti = mti + 1
 y = mt[mti]

 y = bit_xor(y, shift_r(y, 11))
 y = bit_xor(y, bit_and(shift_l(y, 7), TEMPERING_MASK_B))
 y = bit_xor(y, bit_and(shift_l(y, 15), TEMPERING_MASK_C))
 y = bit_xor(y, shift_r(y, 18))
 
 return y
end function

-- just like set_rand() .. seeds the twister
global procedure mt_set_rand(integer new_seed)
 sgenrand(new_seed)
end procedure

atom the_ret
atom the_rand_num

-- just like rand() .. returns a number from 1 to max_rand inclusive
global function mt_rand(integer max_rand)
 
 the_rand_num = genrand()
 if (the_rand_num > max_rand) then
  the_ret = remainder(the_rand_num, max_rand) + 1
 else
  the_ret = the_rand_num
 end if
 -- not sure if this'd happen anyway .. doesn't hurt to be sure
 if (the_ret = 0) then
  the_ret = max_rand -- instead of 1
 end if 
 return the_ret
end function

-- just like mt_rand() .. limited to #FFFFFFFF instead of #3FFFFFFF
global function mt_rand_a(atom max_rand)
 
 the_rand_num = genrand()
 if (the_rand_num > max_rand) then
  the_ret = remainder(the_rand_num, max_rand) + 1
 else
  the_ret = the_rand_num
 end if
 -- not sure if this'd happen anyway .. doesn't hurt to be sure
 if (the_ret = 0) then
  the_ret = max_rand -- instead of 1
 end if 
 return the_ret
end function

-- mt_get() and mt_set(s, i) are used to save and restore the state of the twister
global procedure mt_set(sequence twister_data, integer twister_index)
 if ((length(twister_data) = N) and (twister_index <= N)) then
  mt = twister_data
  mti = twister_index
 end if
end procedure

global function mt_get()
 return {mt, mti}
end function

